/*
 * GWT - General Windowing Toolkit
 *
 * Copyright (C) 1998 Tristan Wibberley - bloater@ps.cus.umist.ac.uk
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Log: draw.c,v $
 * Revision 1.12  2000/06/09 11:34:44  ortalo
 * Added support for drawing functions based on LibXMI.
 *
 * Revision 1.11  2000/03/19 19:32:46  ortalo
 * Some debugging and modifications (wrt OpenAmulet mainly).
 *
 * Revision 1.10  2000/03/13 23:05:41  ortalo
 * Corrected a translation bug in gwt{Get,Set}Clippingi wrt local/global
 * coordinates.
 *
 * Revision 1.9  2000/03/13 22:02:55  ortalo
 * Committed new version of LibGWT corresponding to 0.0.2.
 * The event management API is probably more or less defined. The simplest
 * approach finally won. (I should have remembered: KISS!)
 *
 * Revision 1.8  1999/05/20 20:34:22  ortalo
 * Several bugs corrections and small improvements.
 *
 * Revision 1.7  1999/05/13 20:01:38  ortalo
 * New version of libgwt. Now using LibGII for input management.
 * Some drawing functions implemented. Visibility management should
 * be rather okay. Still some work to do, but now, your eyes can
 * look at some drawings... not only code. ;-)
 *
 * Revision 1.6  1998/11/17 15:13:27  tristan
 *
 * Changed my email address
 *
 * Revision 1.5  1998/09/02 02:21:43  tristan
 * Loadsa great stuff done. Visible/invalid regions,
 *  copy on move - fixed tt_region.c too!
 *
 * Revision 1.4  1998/08/16 18:40:53  ortalo
 * A new version, STILL IN VERY ALPHA STATE... (Be patient.)
 * (However, this version is a little better... :-)
 *
 * Revision 1.3  1998/08/15 16:15:29  marcus
 * Fixed to compile with -pedantic
 *
 * Revision 1.2  1998/08/11 22:17:32  ortalo
 * Adapted to the gwt.h->ggi/gwt.h change. Incorporated the modifications
 * brought by Tristan Wibberley.
 *
 * Revision 1.1  1998/08/11 22:10:06  ortalo
 * Incorporated the changes made by Tristan Wibberley. The code is
 * still in alpha development state.
 *
 */

#include "internal.h"

/************
 * Internal *
 ************/
static gwt_window_t last_prepared = NULL;
static void prepare_for_drawing(gwt_window_t win)
{
  if (win != last_prepared)
    {
      ggiSetGCForeground(win->visual, win->foreground);
      ggiSetGCBackground(win->visual, win->background);
      last_prepared = win;
    }
  return;
}
static void window_gc_changed(gwt_window_t win)
{
  if (last_prepared == win)
    last_prepared = NULL;
  return;
}
/**************************************
 *    GWT internal update procedure   *
 **************************************/
void gwt_drawing_update(const gwt_window_t win)
{
  gwt_region_copy(win->drawing, win->visible);
  if (win->clipping != NULL)
    gwt_region_intersection(win->drawing, win->clipping);
  return;
}
/***********************************
 * Window-specific graphic context *
 ***********************************/
int gwtGetGCForeground(gwt_window_t win, ggi_pixel* color)
{
  *color = win->foreground;
  return 0;
}
int gwtSetGCForeground(gwt_window_t win, ggi_pixel color)
{
  win->foreground = color;
  window_gc_changed(win);
  return 0;
}
int gwtGetGCBackground(gwt_window_t win, ggi_pixel* color)
{
  *color = win->background;
  return 0;
}
int gwtSetGCBackground(gwt_window_t win, ggi_pixel color)
{
  win->background = color;
  window_gc_changed(win);
  return 0;
}
int gwtSetClipping(gwt_window_t win, gwt_region_t clip)
{
  if (clip == NULL)
    {
      gwt_region_destroy(win->clipping);
      win->clipping = NULL;
    }
  else
    {
      ggi_coord geom_tl, geom_br;
      if (win->clipping == NULL)
	gwt_region_create(&(win->clipping));
      gwt_region_copy(win->clipping, clip);
      /* Moves the clipping region to root window coordinates */
      gwt_region_bounds(win->geometry, &geom_tl, &geom_br);
      gwt_region_move(win->clipping, geom_tl.x, geom_tl.y);
    }
  gwt_drawing_update(win);
  return 0;
}
int gwtGetClipping(gwt_window_t win, gwt_region_t clip)
{
  ggi_coord geom_tl, geom_br;
  gwt_region_copy(clip, win->clipping);
  /* Moves the clipping region back window local coordinates */
  gwt_region_bounds(win->geometry, &geom_tl, &geom_br);
  gwt_region_move(clip, -geom_tl.x, -geom_tl.y);
  return 0;
}
/*******************
 * Window flushing *
 *******************/
int gwtFlush(gwt_window_t win)
{
  ggi_coord glob_tl, glob_br;
  if (!gwtIsVisible(win))
    return -1;
  /* We flush the geometry bounding box... */
  gwt_region_bounds(win->geometry, &glob_tl, &glob_br);
  ggiFlushRegion(win->visual, glob_tl.x, glob_tl.y,
		 glob_br.x - glob_tl.x, glob_br.y - glob_tl.y);
  return 0;
}
/******************
 *  Pixel drawing *
 ******************/
/*
 * Draws a pixel
 */
int gwtDrawPixel(gwt_window_t win, int x, int y)
{
  if (!gwtIsVisible(win))
    return -1;
  prepare_for_drawing(win);
  x += win->origin.x; y += win->origin.y;
  if (gwt_region_contains_point(win->drawing, x, y))
    return ggiDrawPixel(win->visual, x, y);
  return 0;
}
/*
 * Puts a pixel
 */
int gwtPutPixel(gwt_window_t win, int x, int y, ggi_pixel pixel)
{
  if (!gwtIsVisible(win))
    return -1;
  prepare_for_drawing(win);
  x += win->origin.x; y += win->origin.y;
  if (gwt_region_contains_point(win->drawing, x, y))
    return ggiPutPixel(win->visual, x, y, pixel);
  return 0;
}

/****************
 * Line drawing *
 ****************/
int gwtDrawLine(gwt_window_t win,int x,int y,int xe,int ye)
{
  int rect_no = 0;
  ggi_coord tl, br, glob_tl, glob_br;
  int t,l,r,b;

  if (!gwtIsVisible(win)) return -1;
  prepare_for_drawing(win);
  gwt_region_bounds(win->geometry, &glob_tl, &glob_br);
  /* TODO: Generalize this save of clipping */
  ggiGetGCClipping(win->visual,&t,&l,&r,&b);
  x += glob_tl.x; y += glob_tl.y;
  xe += glob_tl.x; ye += glob_tl.y;
  while (gwt_region_rect(win->drawing, rect_no++, &tl, &br))
    {
      ggiSetGCClipping(win->visual,tl.x,tl.y,br.x,br.y);
      ggiDrawLine(win->visual,x,y,xe,ye);
    }
  /* TODO: Generalize this save of clipping */
  ggiSetGCClipping(win->visual,t,l,r,b);
  return 0;
}
#if 0
/* TODO: Implement */
  int gwtDrawHLine(gwt_window_t win,int x,int y,int w);
  int gwtDrawVLine(gwt_window_t win,int x,int y,int h);
#endif

/***************
 * Box drawing *
 ***************/
int gwtDrawBox(gwt_window_t win,int x,int y,int w,int h)
{
  int rect_no = 0;
  ggi_coord tl, br, glob_tl, glob_br;
  int t,l,r,b;

  if (!gwtIsVisible(win)) return -1;
  prepare_for_drawing(win);
  gwt_region_bounds(win->geometry, &glob_tl, &glob_br);
  /* TODO: Generalize this save of clipping */
  ggiGetGCClipping(win->visual,&t,&l,&r,&b);
  x += glob_tl.x; y += glob_tl.y;
  while (gwt_region_rect(win->drawing, rect_no++, &tl, &br))
    {
      ggiSetGCClipping(win->visual,tl.x,tl.y,br.x,br.y);
      ggiDrawBox(win->visual,x,y,w,h);
    }
  /* TODO: Generalize this save of clipping */
  ggiSetGCClipping(win->visual,t,l,r,b);
  return 0;
}

/***************
 * Cross Blits *
 ***************/
int gwtCrossBlit(gwt_window_t src, int sx, int sy, int sw, int sh,
		 gwt_window_t dst, int dx, int dy)
{
  int rect_no = 0;
  ggi_coord tl, br, src_tl, src_br, dst_tl, dst_br;
  int t,l,r,b;

  if (!gwtIsVisible(dst)) return -1;
  prepare_for_drawing(dst);
  gwt_region_bounds(src->geometry, &src_tl, &src_br);
  gwt_region_bounds(dst->geometry, &dst_tl, &dst_br);
  /* TODO: Generalize this save of clipping */
  ggiGetGCClipping(dst->visual,&t,&l,&r,&b);
  sx += src_tl.x; sy += src_tl.y;
  dx += dst_tl.x; dy += dst_tl.y;
  /* TODO: Check if we do not need to take into account the
     TODO: visible state of the _source_ window also !!!
     TODO: (And maybe send expose events or sthing like this.) */
  while (gwt_region_rect(dst->drawing, rect_no++, &tl, &br))
    {
      ggiSetGCClipping(dst->visual,tl.x,tl.y,br.x,br.y);
      ggiCrossBlit(src->visual,sx,sy,sw,sh,
		   dst->visual,dx,dy);
    }
  /* TODO: Generalize this save of clipping */
  ggiSetGCClipping(dst->visual,t,l,r,b);
  return 0;
}

/****************
 * Char drawing *
 ****************/
int gwtPutc(gwt_window_t win,int x,int y,char c)
{
  int rect_no = 0;
  ggi_coord tl, br, glob_tl, glob_br;
  int t,l,r,b;

  if (!gwtIsVisible(win)) return -1;
  prepare_for_drawing(win);
  gwt_region_bounds(win->geometry, &glob_tl, &glob_br);
  /* TODO: Generalize this save of clipping */
  ggiGetGCClipping(win->visual,&t,&l,&r,&b);
  x += glob_tl.x; y += glob_tl.y;
  while (gwt_region_rect(win->drawing, rect_no++, &tl, &br))
    {
      ggiSetGCClipping(win->visual,tl.x,tl.y,br.x,br.y);
      ggiPutc(win->visual,x,y,c);
    }
  /* TODO: Generalize this save of clipping */
  ggiSetGCClipping(win->visual,t,l,r,b);
  return 0;
}
int gwtPuts(gwt_window_t win,int x,int y,const char *str)
{
  int rect_no = 0;
  ggi_coord tl, br, glob_tl, glob_br;
  int t,l,r,b;

  /*  printf("Drawing this string at (%i,%i): '%s'\n", x, y, str); */

  if (!gwtIsVisible(win)) return -1;
  prepare_for_drawing(win);
  gwt_region_bounds(win->geometry, &glob_tl, &glob_br);
  /* TODO: Generalize this save of clipping */
  ggiGetGCClipping(win->visual,&t,&l,&r,&b);
  x += glob_tl.x; y += glob_tl.y;
  while (gwt_region_rect(win->drawing, rect_no++, &tl, &br))
    {
      ggiSetGCClipping(win->visual,tl.x,tl.y,br.x,br.y);
      /* printf(" = first clipping the string at (%i,%i) (%i,%i)\n",tl.x,tl.y,br.x,br.y); */
      ggiPuts(win->visual,x,y,str);
      /* printf(" = then drawing this string at (%i,%i): '%s'\n", x, y, str); */
    }
  /* TODO: Generalize this save of clipping */
  ggiSetGCClipping(win->visual,t,l,r,b);
  return 0;
}
/******************
 *  Miscellaneous *
 ******************/
/*
 * Draws the rectangular window outline (mainly for debugging)
 */
int gwtDrawFrame(gwt_window_t win)
{
  ggi_coord tl, br;
  
  if (!gwtIsVisible(win)) return -1;
  prepare_for_drawing(win);
  gwt_region_bounds(win->geometry, &tl, &br);
  ggiDrawHLine(win->visual, tl.x,   tl.y,   br.x-tl.x);
  ggiDrawHLine(win->visual, tl.x,   br.y-1, br.x-tl.x);
  ggiDrawVLine(win->visual, tl.x,   tl.y,   br.y-tl.y);
  ggiDrawVLine(win->visual, br.x-1, tl.y,   br.y-tl.y);
  return 0;
}
/* These next two functions can be altered to operate on the invalid
 * region instead of the visible one. Just change win->visible to
 * win->invalid :)
 */
int gwtDrawVisibleRegion(gwt_window_t win)
{
  int rect_no = 0;
  ggi_coord tl, br;

  if (!gwtIsVisible(win))
    return -1;
  prepare_for_drawing(win);
  while (gwt_region_rect(win->visible, rect_no++, &tl, &br))
    {
      ggiDrawHLine(win->visual, tl.x,   tl.y,   br.x-tl.x);
      ggiDrawHLine(win->visual, tl.x,   br.y-1, br.x-tl.x);
      ggiDrawVLine(win->visual, tl.x,   tl.y,   br.y-tl.y);
      ggiDrawVLine(win->visual, br.x-1, tl.y,   br.y-tl.y);
    }
  return 0;
}
int gwtDrawInvalidRegion(gwt_window_t win)
{
  int rect_no = 0;
  ggi_coord tl, br;

  if (!gwtIsVisible(win))
    return -1;
  prepare_for_drawing(win);
  while (gwt_region_rect(win->invalid, rect_no++, &tl, &br))
    {
      ggiDrawHLine(win->visual, tl.x,   tl.y,   br.x-tl.x);
      ggiDrawHLine(win->visual, tl.x,   br.y-1, br.x-tl.x);
      ggiDrawVLine(win->visual, tl.x,   tl.y,   br.y-tl.y);
      ggiDrawVLine(win->visual, br.x-1, tl.y,   br.y-tl.y);
    }
  return 0;
}
int gwtFillscreen(gwt_window_t win)
{
  int rect_no=0;
  int t,l,r,b;
  ggi_coord tl, br;
  
  if (!gwtIsVisible(win)) return -1;
  prepare_for_drawing(win);
  /* TODO: Generalize this save of clipping */
  ggiGetGCClipping(win->visual,&t,&l,&r,&b);
  while (gwt_region_rect(win->visible, rect_no++, &tl, &br))
    {
      ggiSetGCClipping(win->visual,tl.x,tl.y,br.x,br.y);
      ggiDrawBox(win->visual, tl.x, tl.y, br.x-tl.x, br.y-tl.y);
    }
  /* TODO: Generalize this save of clipping */
  ggiSetGCClipping(win->visual,t,l,r,b);
  return 0;
}

int gwtCopyBox(gwt_window_t win, int x, int y, int w, int h, int nx, int ny)
{
  gwt_region_t box, boxint;
  int t,l,r,b;
  ggi_coord tl, br;
  int rect_no=0;
  
  if (!gwtIsVisible(win)) return -1;
  gwt_region_create(&box);
  gwt_region_create(&boxint);
  gwt_region_init(box, x+win->origin.x, y+win->origin.y, x+w+win->origin.x, y+h+win->origin.y);
  gwt_region_copy(boxint, win->drawing);
  gwt_region_intersection(boxint, box);
  /* TODO: Generalize this save of clipping */
  ggiGetGCClipping(win->visual,&t,&l,&r,&b);
  while (gwt_region_rect(win->visible, rect_no++, &tl, &br))
    {
      ggiSetGCClipping(win->visual,tl.x,tl.y,br.x,br.y);
      ggiCopyBox(win->visual, tl.x, tl.y, br.x-tl.x, br.y-tl.y, nx+win->origin.x, ny+win->origin.y);
    }
  /* TODO: Generalize this save of clipping */
  ggiSetGCClipping(win->visual,t,l,r,b);
  return 0;
}

int gwt_copy_displayed_region(ggi_visual_t vis, gwt_region_t region, int dx, int dy)
{
	ggi_coord tl, br;
	int rect_no=0;
	while (gwt_region_rect(region, rect_no++, &tl, &br))
		ggiCopyBox(vis, tl.x, tl.y, br.x-tl.x, br.y-tl.y, tl.x+dx, tl.y+dy);
	return 0;
}

/****************************
 * LibXMI wrapper functions *
 ****************************/
/*
 * Most of these functions are pure wrappers
 */
miPaintedSet * gwtmiNewPaintedSet(gwt_window_t win)
{
  return miNewPaintedSet(win->visual);
}
void gwtmiDeletePaintedSet(gwt_window_t win, miPaintedSet *paintedSet)
{
  miDeletePaintedSet(win->visual, paintedSet);
}
void gwtmiClearPaintedSet(gwt_window_t win, miPaintedSet *paintedSet)
{
  miClearPaintedSet(win->visual, paintedSet);
}
miGC * gwtmiNewGC(gwt_window_t win, int npixels, const miPixel *pixels,
		    int nblendstages, const miBlendStage *blendstages)
{
  return miNewGC(win->visual, npixels, pixels, nblendstages, blendstages);
}
void gwtmiDeleteGC(gwt_window_t win, miGC *pGC)
{
  miDeleteGC(win->visual, pGC);
}
miGC * gwtmiCopyGC(gwt_window_t win, const miGC *pGC)
{
  return miCopyGC(win->visual, pGC);
}
void gwtmiSetGCAttrib(gwt_window_t win, miGC *pGC, miGCAttribute attribute, int value)
{
  miSetGCAttrib(win->visual, pGC, attribute, value);
}
void gwtmiSetGCAttribs(gwt_window_t win, miGC *pGC, int nattributes,
		       const miGCAttribute *attributes, const int *values)
{
  /* NB: NIY  miSetGCAttribs(win->visual, pGC, nattributes, attributes, values); */
}
void gwtmiSetGCDashes(gwt_window_t win, miGC *pGC, int ndashes, const unsigned int *dashes, int offset)
{
  miSetGCDashes(win->visual, pGC, ndashes, dashes, offset);
}
void gwtmiSetGCMiterLimit(gwt_window_t win, miGC *pGC, double miter_limit)
{
  /* NB: NIY  miSetGCMiterLimit(win->visual, pGC, miter_limit); */
}
void gwtmiSetGCPixels(gwt_window_t win, miGC *pGC, int npixels, const miPixel *pixels)
{
  /* NB: NIY  miSetGCPixels(win->visual, pGC, npixels, pixels); */
}
void gwtmiDrawPoints(gwt_window_t win, miPaintedSet *paintedSet, const miGC *pGC, miCoordMode mode,
		     int npts, const miPoint *pPts)
{
  miDrawPoints(win->visual, paintedSet, pGC, mode, npts, pPts);
}
void gwtmiDrawLines(gwt_window_t win, miPaintedSet *paintedSet, const miGC *pGC, miCoordMode mode,
		    int npts, const miPoint *pPts)
{
  miDrawLines(win->visual, paintedSet, pGC, mode, npts, pPts);
}
void gwtmiFillPolygon(gwt_window_t win, miPaintedSet *paintedSet, const miGC *pGC,
		      miPolygonShape shape, miCoordMode mode, int npts, const miPoint *pPts)
{
  miFillPolygon(win->visual, paintedSet, pGC, shape, mode, npts, pPts);
}
void gwtmiDrawRectangles(gwt_window_t win, miPaintedSet *paintedSet, const miGC *pGC,
			 int nrects, const miRectangle *pRects)
{
  miDrawRectangles(win->visual, paintedSet, pGC, nrects, pRects);
}
void gwtmiFillRectangles(gwt_window_t win, miPaintedSet *paintedSet, const miGC *pGC,
			 int nrects, const miRectangle *pRects)
{
  miFillRectangles(win->visual, paintedSet, pGC, nrects, pRects);
}
void gwtmiDrawArcs(gwt_window_t win, miPaintedSet *paintedSet, const miGC *pGC, int narcs, const miArc *parcs)
{
  miDrawArcs(win->visual, paintedSet, pGC, narcs, parcs);
}
void gwtmiFillArcs(gwt_window_t win, miPaintedSet *paintedSet, const miGC *pGC, int narcs, const miArc *parcs)
{
  miFillArcs(win->visual, paintedSet, pGC, narcs, parcs);
}
miPixmap * gwtmiNewPixmap(uint w, uint h, ggi_graphtype type)
{
  return miNewPixmap(w, h, type);
}
void gwtmiDeletePixmap(gwt_window_t win, miPixmap *pPixmap)
{
  miDeletePixmap(win->visual, pPixmap);
}
/*
 * This functions does the actual copying to the root visual:
 * we manage the clipping to the target window then.
 */
void gwtmiCopyPaintedSetToVisual(gwt_window_t win, const miGC *pGC,
				   const miPaintedSet *paintedSet, miPoint offset)
{
  int rect_no = 0;
  ggi_coord tl, br, glob_tl, glob_br;
  int t,l,r,b;
  
  if (!gwtIsVisible(win)) return;
  prepare_for_drawing(win);
  gwt_region_bounds(win->geometry, &glob_tl, &glob_br);
  /* TODO: Generalize this save of clipping */
  ggiGetGCClipping(win->visual,&t,&l,&r,&b);
  offset.x += glob_tl.x; offset.y += glob_tl.y;
  while (gwt_region_rect(win->drawing, rect_no++, &tl, &br))
    {
      ggiSetGCClipping(win->visual,tl.x,tl.y,br.x,br.y);
      miCopyPaintedSetToVisual(win->visual, pGC, paintedSet, offset);
    }
  /* TODO: Generalize this save of clipping */
  ggiSetGCClipping(win->visual,t,l,r,b);
}
